{-# OPTIONS_GHC -fth -cpp #-}

{-|
  Module      :  RecordUpdate

  Maintainer  :  Misha Aizatulin <avatar@hot.ee>
  Portability :  ghc (Template Haskell)

  Automatic generation of record updaters.
-} 

module Lib.RecordUpdate (genUpdaters, genFieldUpdaters) where

import Lib.TypeInfo
import Lib.Misc (updateIndex)

import Language.Haskell.TH

import Data.Char (toUpper)
import Data.Maybe (fromJust)
import Control.Monad (liftM)

{-|
  Given a record type @Rec@, generate for each field @fieldName :: T@ the updater 

  > updateFieldName :: (T -> T) -> Rec -> Rec

  Currently can only be used with haskell98 types, because other types cannot be reified
  with template haskell. You can still use 'genFieldUpdaters' for other types.
-}
genUpdaters :: Name -> Q [Dec]
genUpdaters name = do
  (_, _, cons) <- typeInfo name
  liftM concat $ sequence $ do
    (con, fields) <- cons
    return $ genFieldUpdaters con $ map (fromJust . fst) fields

-- the version used with old genUpdater:
-- genUpdaters :: Name -> Q [Dec]
-- genUpdaters name = do
--   (_, _, cons) <- typeInfo name
--   mapM genUpdater [field | (_, fields) <- cons, (Just field, _) <- fields]


{-|
  Works like 'genUpdaters', but takes the constructor name and the list of 
  fields, which should be complete. This function can be useful for 
  non-haskell98 types that presently cannot be used with 'genUpdaters'.
-}
genFieldUpdaters :: Name -> [Name] -> Q [Dec]
genFieldUpdaters con fields = sequence $ do
  i <- [0..length fields - 1]
  return $ genUpdater con (length fields) i (fields !! i)


genUpdater :: Name {-^ constructor name -} 
            -> Int {-^ constructor arity -}  
            -> Int {-^ field number -} 
            -> Name {-^ field name -} 
            -> Q Dec
#ifndef __HADDOCK__ 
genUpdater con arity fieldNum field = do
  fields <- mapM newName $ take arity $ repeat "a"
  f      <- newName "f"
  let
    funName = mkName $ "update" ++ (capitalize $ nameBase field)
    fields' = updateIndex fieldNum ((varE f) `appE`) $ map varE fields
    fun = normalB (appsE $ (conE con):(fields'))
  funD funName $ [clause [varP f, conP con (map varP fields)] fun []]
#endif 
-- __HADDOCK__


-- The old version using record updates. These cannot be used for non-haskell98 types yet.

-- genUpdater :: Name -> Q Dec
--- #ifndef __HADDOCK__ 
-- genUpdater name = do
--   let 
--     funName = mkName $ "update" ++ (capitalize $ nameBase name)
--     update f a = do 
--       e <- f `appE` (varE name `appE` a)
--       return (name, e)
--     fun = normalB [| (\f a -> $(recUpdE [|a|] [update [|f|] [|a|]])) |]
--   funD funName $ [clause [] fun []]
-- -- #endif 
--- __HADDOCK__


capitalize :: String -> String
capitalize (c:cs) = toUpper c : cs
capitalize []     = []
